package cz.mendelu.dnaAnalyser.analyse.cpg;


import lombok.Getter;

public class CpgWindow {

    public final String record;
    private final int recordLen;
    public int windowBegin;
    public int windowSize;
    public int windowEnd;
    private int gcCount;
    private double obsCpg;
    @Getter
    private double obsExp;
    @Getter
    private double gcPerc;
    private final double MIN_GC_PERCENTAGE;
    private final double MIN_OBSERVED_TO_EXPECTED_CPG;
    private final int MIN_WINDOW_SIZE;
    private final char FIRST_NUCLEOTIDE;
    private final char SECOND_NUCLEOTIDE;

    public CpgWindow(   String sequence,
                        int windowBegin,
                        int windowSize,
                        double MIN_GC_PERCENTAGE,
                        double MIN_OBSERVED_TO_EXPECTED_CPG,
                        int MIN_WINDOW_SIZE,
                        char FIRST_NUCLEOTIDE,
                        char SECOND_NUCLEOTIDE) {

        this.record = sequence;
        this.windowBegin = windowBegin;
        this.windowSize = windowSize;
        this.recordLen = sequence.length();
        this.windowEnd = windowBegin + windowSize;
        this.gcCount = 0;
        this.obsCpg = 0.0;
        this.obsExp = 0.0;
        this.gcPerc = 0.0;
        this.MIN_GC_PERCENTAGE = MIN_GC_PERCENTAGE;
        this.MIN_OBSERVED_TO_EXPECTED_CPG = MIN_OBSERVED_TO_EXPECTED_CPG;
        this.MIN_WINDOW_SIZE = MIN_WINDOW_SIZE;
        this.FIRST_NUCLEOTIDE = FIRST_NUCLEOTIDE;
        this.SECOND_NUCLEOTIDE = SECOND_NUCLEOTIDE;
        this.updateGcCount();

    }


    public String toString() {
        return "Window at (" + this.windowBegin + ", " + this.windowEnd + ")\tgcper:" + this.gcPerc + "\tobs_exp:" + this.obsExp + "\twin_length:" + this.windowSize;
    }

    private void evaluate() {
        this.gcPerc = (double) this.gcCount / this.windowSize;
        double expCpg = Math.pow(((double) this.gcCount / 2), 2) / this.windowSize;
        try {
            this.obsExp = this.obsCpg / expCpg;
        } catch (ArithmeticException e) {
            this.obsExp = 0.0;
        }
    }

    public boolean isIsland() {
        this.evaluate();
        return this.gcPerc > MIN_GC_PERCENTAGE && this.obsExp > MIN_OBSERVED_TO_EXPECTED_CPG;
    }

    private int count(String seq, String subsequence) {
        int count = 0;
        for (int i = 0; i < seq.length(); i++) {
            if (seq.charAt(i) == subsequence.charAt(0)) {
                if (seq.startsWith(subsequence, i)) {
                    count++;
                }
            }
        }
        return count;

    }

    public void updateGcCount() {
        String subsequence = this.record.substring(this.windowBegin, this.windowEnd);
        this.gcCount = this.count(subsequence, String.valueOf(FIRST_NUCLEOTIDE)) + count(subsequence, String.valueOf(SECOND_NUCLEOTIDE));
        this.obsCpg = this.count(subsequence,String.valueOf(FIRST_NUCLEOTIDE) + String.valueOf(SECOND_NUCLEOTIDE));
        this.evaluate();
    }

    public void shrinkLeft() {
        // check if window begin is First or second nucleotide
        if (charAt(this.record, this.windowBegin) == FIRST_NUCLEOTIDE || charAt(this.record, this.windowBegin) == SECOND_NUCLEOTIDE) {
            this.gcCount--;
        }
        if (charAt(this.record, this.windowBegin) == FIRST_NUCLEOTIDE && charAt(this.record, this.windowBegin + 1) == SECOND_NUCLEOTIDE) {
            this.obsCpg--;
        }
        this.windowBegin++;
    }


    public boolean expandRight() {
        if (this.windowEnd >= this.recordLen) {
            return false;
        }
        if (charAt(this.record, this.windowEnd) == FIRST_NUCLEOTIDE || charAt(this.record, this.windowEnd) == SECOND_NUCLEOTIDE) {
            this.gcCount++;
        }
        if (charAt(this.record, this.windowEnd - 1) == FIRST_NUCLEOTIDE && charAt(this.record, this.windowEnd) == SECOND_NUCLEOTIDE) {
            this.obsCpg++;
        }
        this.windowEnd++;
        return true;
    }


    public void shrinkRight() {
        if (charAt(this.record, this.windowEnd - 1) == FIRST_NUCLEOTIDE || charAt(this.record, this.windowEnd - 1) == SECOND_NUCLEOTIDE) {
            this.gcCount--;
        }
        if (charAt(this.record, this.windowEnd - 2) == FIRST_NUCLEOTIDE && charAt(this.record, this.windowEnd - 1) == SECOND_NUCLEOTIDE) {
            this.obsCpg--;
        }
        this.windowEnd--;
    }

    private void expandLeft() {
        if (charAt(this.record, this.windowBegin - 1) == FIRST_NUCLEOTIDE || charAt(this.record, this.windowBegin - 1) == SECOND_NUCLEOTIDE) {
            this.gcCount++;
        }
        if (charAt(this.record, this.windowBegin - 2) == FIRST_NUCLEOTIDE && charAt(this.record, this.windowBegin - 1) == SECOND_NUCLEOTIDE) {
            this.obsCpg++;
        }
        this.windowBegin--;
    }

    public boolean expandRightIsland() {
        int remaining = this.recordLen - this.windowEnd;
        int newWindowBegin;
        if (remaining == 0) {
            return false;
        }
        if (remaining < MIN_WINDOW_SIZE) {
            newWindowBegin = this.recordLen - MIN_WINDOW_SIZE;
        } else {
            newWindowBegin = this.windowEnd;
        }
        CpgWindow newWindow = new CpgWindow(this.record, newWindowBegin, this.MIN_WINDOW_SIZE, this.MIN_GC_PERCENTAGE, this.MIN_OBSERVED_TO_EXPECTED_CPG, this.MIN_WINDOW_SIZE, this.FIRST_NUCLEOTIDE, this.SECOND_NUCLEOTIDE);
        this.windowEnd = newWindowBegin + this.MIN_WINDOW_SIZE;
        this.windowSize = this.windowEnd - this.windowBegin;
        return newWindow.isIsland();
    }

    public void rollbackUntilIsland() {
        CpgWindow lastWindow = new CpgWindow(this.record, this.windowEnd - this.MIN_WINDOW_SIZE, this.MIN_WINDOW_SIZE, this.MIN_GC_PERCENTAGE, this.MIN_OBSERVED_TO_EXPECTED_CPG, this.MIN_WINDOW_SIZE, this.FIRST_NUCLEOTIDE, this.SECOND_NUCLEOTIDE);
        while (!lastWindow.isIsland()) {
            lastWindow.expandLeft();
            lastWindow.shrinkRight();
        }
        this.joinWindows(lastWindow);
    }

    public void joinWindows(CpgWindow lastWindow) {
        this.windowEnd = lastWindow.windowEnd;
        this.windowSize = this.windowEnd - this.windowBegin;
        this.updateGcCount();
    }

    public static char charAt(String str, int index) {
        if (str == null) {
            throw new NullPointerException("String is null");
        }

        // If the index is negative, adjust it to access the character from the end of the string
        if (index < 0) {
            index += str.length();
        }

        // Check if the adjusted index is still out of range
        if (index < 0 || index >= str.length()) {
            throw new StringIndexOutOfBoundsException("String index out of range: " + index);
        }

        return str.charAt(index);
    }
}
