package cz.mendelu.dnaAnalyser.sequence

import cz.mendelu.dnaAnalyser.sequence.inport.ImportSequenceFormat
import cz.mendelu.dnaAnalyser.sequence.inport.file.ImportSequenceFromFileRequest
import cz.mendelu.dnaAnalyser.sequence.inport.text.ImportSequenceFromTextRequest
import cz.mendelu.dnaAnalyser.utils.api.ApiRestTemplate
import cz.mendelu.dnaAnalyser.utils.controller.TagModifyRequest
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.boot.test.context.SpringBootTest
import org.springframework.http.HttpStatus
import org.springframework.http.MediaType
import org.springframework.test.context.ActiveProfiles
import spock.lang.Ignore
import spock.lang.Specification

import static org.springframework.boot.test.context.SpringBootTest.WebEnvironment.RANDOM_PORT

@Ignore
@ActiveProfiles("test")
@SpringBootTest(webEnvironment = RANDOM_PORT)
class SequenceControllerTest extends Specification {

    @Autowired
    ApiRestTemplate api

    def setup() {
        api.testAuthentication('SequenceControllerTest')
    }

    def "GET:/api/sequence Get page with sequences"() {
        setup:
        api.authentication('user@mendelu.cz', "user")

        when:
        def response = api.get("/api/sequence")

        then:
        response.status == HttpStatus.OK.value()

        and:
        def items = response.body.items
        items.size == 2

        and:
        def page = response.body.page
        page.pageSize == 20
        page.pageNumber == 0
        page.totalPages == 1
        page.totalElements == 2
    }

    def "GET:/api/sequence/tag Get all tag on sequences"() {
        setup:
        api.authentication('user@mendelu.cz', "user")

        when:
        def response = api.get("/api/sequence/tag")

        then:
        response.status == HttpStatus.OK.value()

        and:
        def items = response.body.items
        items.size == 2

    }

    def "GET:/api/sequence/{id} Get one sequence by ID"() {
        setup:
        api.authentication('user@mendelu.cz', "user")
        def id = UUID.fromString('987acc67-5714-47b1-b56f-1dc56ded7c87')

        when:
        def response = api.get("/api/sequence/${id}")

        then:
        response.status == HttpStatus.OK.value()

        and:
        def payload = response.body.payload
        payload.id == '987acc67-5714-47b1-b56f-1dc56ded7c87'
        payload.name == 'Escherichia coli str. K-12 substr. MG1655'
    }

    def "GET:/api/sequence/{id} Get one sequence by ID owned by another user"() {
        setup:
        api.authentication('user@mendelu.cz', "user")
        def id = UUID.fromString('5f9a8c63-71da-47c5-9130-31863476de9d')

        when:
        def response = api.get("/api/sequence/${id}")

        then:
        response.status == HttpStatus.NOT_FOUND.value()
    }

    @Ignore
    def "GET:/api/sequence/{id}/data Get sequence data"() {
        setup:
        // Copy sequence form src/resources/sequences to ./dev/sequences
        api.authentication('user@mendelu.cz', "user")
        def id = UUID.fromString('987acc67-5714-47b1-b56f-1dc56ded7c87')

        when:
        def response = api.get("/api/sequence/${id}/data", MediaType.TEXT_PLAIN)

        then:
        response.status == HttpStatus.OK.value()
        response.body.length == 100
    }

    def "DELETE:/api/sequence/{id} Delete one sequence by ID"() {
        setup:
        api.authentication('admin@mendelu.cz', "admin")
        def id = UUID.fromString('95db298b-5209-4611-90ad-0f601e7b53fd')

        when:
        def response = api.delete("/api/sequence/${id}")

        then:
        response.status == HttpStatus.NO_CONTENT.value()
    }

    def "DELETE:/api/sequence/{id} Delete one sequence by ID owned by another user"() {
        setup:
        api.authentication('user@mendelu.cz', "user")
        def id = UUID.fromString('5f9a8c63-71da-47c5-9130-31863476de9d')

        when:
        def response = api.delete("/api/sequence/${id}")

        then:
        response.status == HttpStatus.NOT_FOUND.value()
    }

    def "POST:/api/sequence/import/text Create sequence from plain text."() {
        given:
        def request = new ImportSequenceFromTextRequest(
                name: "ImportSequenceFromTextRequest",
                format: ImportSequenceFormat.PLAIN,
                circular: true,
                type: SequenceType.DNA,
                tags: ['test', 'import', 'plain text'],
                data: 'ATCG'
        )

        when:
        def response = api.post("/api/sequence/import/text", request)

        then:
        response.status == HttpStatus.CREATED.value()

        and:
        def sequence = response.body.payload
        sequence.id != null
        sequence.name == "ImportSequenceFromTextRequest"
        //sequence.created != null
        sequence.type == 'DNA'
        sequence.circular == true
        sequence.tags == ['test', 'import', 'plain text']

        //and:
        //def stored = sequenceRepository.findOne(UUID.fromString(sequence.id))
        //stored != null
    }

    @Ignore
    def "POST:/api/sequence/import/file Create sequence from plain file."() {
        given:
        def request = new ImportSequenceFromFileRequest(
                name: "ImportSequenceFromFileRequest",
                format: ImportSequenceFormat.PLAIN,
                circular: false,
                type: SequenceType.DNA,
                tags: ['test', 'import', 'plain file']
        )

        when:
        def response = api.post("/api/sequence/import/file", request)

        then:
        response.status == HttpStatus.CREATED.value()

        and:
        def sequence = response.body.payload
        sequence.id != null
        sequence.name == "ImportSequenceFromFileRequest"
        sequence.created != null
        sequence.type == 'DNA'
        sequence.circular == false
        sequence.tags == ['test', 'import', 'plain file']
    }

    @Ignore
    def "POST:/api/sequence/import/ncbi Post sequence base on NCBI ID."() {
        given:
        def request = new SequenceController.ImportNCBIRequest(
                tags: ['jedna', 'dva'],
                ncbis: [
                        new SequenceController.ImportNCBIRequest.NCBI(name: 'zivocich', ncbiId: 'J01859.1'),
                        new SequenceController.ImportNCBIRequest.NCBI(name: 'rostlina', ncbiId: 'KB944764.1'),
                        new SequenceController.ImportNCBIRequest.NCBI(name: 'obojetnik', ncbiId: 'KB944765.1')
                ]
        )

        when:
        def response = api.post("/api/sequence/import/ncbi", request)

        then:
        response.status == HttpStatus.CREATED.value()
        response.body.size == 3
    }

    @Ignore
    def "PATCH:/api/sequence/{id} Update statistic stored sequence"() {
        setup:
        api.authentication('user@mendelu.cz', "user")
        def id = UUID.fromString('090e5461-6dd3-44ff-8e01-7b84435857fa')
        def request = new SequenceController.PatchSequenceRequest()

        when:
        def response = api.patch("/api/sequence/${id}", request)

        then:
        response.status == HttpStatus.OK.value()

        and:
        def payload = response.body.payload
        payload.nucleicCounts != null
        payload.size == 4
    }

    def "PUT:/api/sequence/{id}/tags Modify tags"() {
        setup:
        api.authentication('user@mendelu.cz', "user")
        def id = UUID.fromString('090e5461-6dd3-44ff-8e01-7b84435857fa')
        def request = new TagModifyRequest(
                tags: ['new tags']
        )

        when:
        def response = api.put("/api/sequence/${id}/tags", request)

        then:
        response.status == HttpStatus.ACCEPTED.value()

        and:
        def payload = response.body.payload
        payload.tags == ['new tags']
    }

}
